# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

import threading
import unittest
import warnings

import six

from instalooter.pbar import ProgressBar, TqdmProgressBar



class TestProgressBar(unittest.TestCase):

    def test_derived_progress_bar(self):

        class MyProgressBar(ProgressBar):
            _test = {"update": 0, "max": None}
            def update(self):
                self._test['update'] += 1
            def set_maximum(self, maximum):
                self._test['max'] = maximum

        pb = MyProgressBar(iter(range(10)))
        self.assertEqual(pb._test['update'], 0)
        self.assertIs(pb._test['max'], None)

        self.assertEqual(next(pb), 0)
        self.assertEqual(pb._test['update'], 1)

        pb.set_maximum(10)
        self.assertEqual(pb._test['max'], 10)

        self.assertEqual(list(pb), list(range(1, 10)))
        self.assertRaises(StopIteration, next, pb)
        self.assertEqual(pb._test['update'], 10)
        pb.finish()

        self.assertRaises(RuntimeError, pb.get_lock)
        lock = threading.RLock()
        pb.set_lock(lock)
        self.assertIs(pb.get_lock(), lock)

    def test_tqdm_progress_bar(self):

        fh = six.moves.StringIO()
        pb = TqdmProgressBar(iter(range(10)), file=fh)

        self.assertEqual(pb.n, 0)
        self.assertIs(pb.total, None)

        self.assertEqual(next(pb), 0)
        self.assertEqual(pb.n, 1)
        self.assertIs(pb.total, None)

        pb.set_maximum(10)
        self.assertEqual(pb.total, 10)

        self.assertEqual(list(pb), list(range(1, 10)))
        self.assertRaises(StopIteration, next, pb)
        self.assertEqual(pb.n, 10)
        pb.finish()

        lock = threading.RLock()
        pb.set_lock(lock)
        self.assertIs(pb.get_lock(), lock)


def setUpModule():
   warnings.simplefilter('ignore')


def tearDownModule():
   warnings.simplefilter(warnings.defaultaction)



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

__author__ = "Martin Larralde"
__author_email__ = "martin.larralde@ens-cachan.fr"
__version__ = "2.3.4"
__license___ = "GPLv3+"



#!/usr/bin/env python
# coding: utf-8
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import functools
import logging
import warnings
import typing

if typing.TYPE_CHECKING:
    from typing import Callable


def warn_logging(logger):
    # type: (logging.Logger) -> Callable
    """Create a `showwarning` function that uses the given logger.

    Arguments:
        logger (~logging.Logger): the logger to use.

    Returns:
        function: a function that can be used as the `warnings.showwarning`
            callback.

    """
    def showwarning(message, category, filename, lineno, file=None, line=None):
        logger.warning(message)
    return showwarning


def wrap_warnings(logger):
    """Have the function patch `warnings.showwarning` with the given logger.

    Arguments:
        logger (~logging.logger): the logger to wrap warnings with when
            the decorated function is called.

    Returns:
        `function`: a decorator function.

    """
    def decorator(func):
        @functools.wraps(func)
        def new_func(*args, **kwargs):
            showwarning = warnings.showwarning
            warnings.showwarning = warn_logging(logger)
            try:
                return func(*args, **kwargs)
            finally:
                warnings.showwarning = showwarning
        return new_func
    return decorator



# coding: utf-8
"""Implementation of the main program executable.

Warning:
    Only `.cli.main` and `.cli.logger` are guaranteed to be stable, do not
    rely on any other member from this package !
"""
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import functools
import logging
import getpass
import os
import sys
import traceback
import warnings

import coloredlogs
import docopt
import fs
import six
import verboselogs

from .. import __version__
from ..looters import InstaLooter, HashtagLooter, ProfileLooter, PostLooter
from ..pbar import TqdmProgressBar
from ..batch import BatchRunner, logger as batch_logger

from . import logutils
from .constants import HELP, USAGE, WARNING_ACTIONS
from .time import get_times_from_cli
from .login import login, logger as login_logger


__all__ = ["main", "logger"]


#: A `~logging.Logger` instance used within the `.cli` module.
logger = verboselogs.VerboseLogger(__name__)


@logutils.wrap_warnings(logger)
def main(argv=None, stream=None):
    """Run from the command line interface.

    Arguments:
        argv (list): The positional arguments to read. Defaults to
            `sys.argv` to use CLI arguments.
        stream (~io.IOBase): A file where to write error messages.
            Leave to `None` to use the `~coloredlogs.StandardErrorHandler`
            for logs, and `sys.stderr` for error messages.

    Returns:
        int: An error code, or 0 if the program executed successfully.
    """

    _print = functools.partial(print, file=stream or sys.stderr)

    # Parse command line arguments
    try:
        args = docopt.docopt(
            HELP, argv, version='instalooter {}'.format(__version__))
    except docopt.DocoptExit as de:
        _print(de)
        return 1

    # Print usage and exit if required (docopt does not do this !)
    if args['--usage']:
        _print(USAGE)
        return 0

    # Set the loggers up with the requested logging level
    level = "ERROR" if args['--quiet'] else args.get("--loglevel", "INFO")
    for logger_ in (logger, login_logger, batch_logger):
        coloredlogs.install(
            level=int(level) if level.isdigit() else level,
            stream=stream,
            logger=logger_)

    # Check the requested logging level
    if args['-W'] not in WARNING_ACTIONS:
        _print("Unknown warning action:", args['-W'])
        _print("    available actions:", ', '.join(WARNING_ACTIONS))
        return 1

    with warnings.catch_warnings():
        warnings.simplefilter(args['-W'])

        try:
            # Run in batch mode
            if args['batch']:
                # Load the batch configuration from the given file
                with open(args['<batch_file>']) as batch_file:
                    batch_runner = BatchRunner(batch_file, args)
                # Run the batch
                batch_runner.run_all()
                return 0

            # Login if requested
            if args['login']:
                try:
                    if not args['--username']:
                        args['--username'] = six.moves.input('Username: ')
                    login(args)
                    return 0
                except ValueError as ve:
                    logger.error("%s", ve)
                    if args["--traceback"]:
                       traceback.print_exc()
                    return 1

            # Logout if requested
            if args['logout']:
                if InstaLooter._cachefs.exists(InstaLooter._COOKIE_FILE):
                    InstaLooter._logout()
                    logger.success('Logged out.')
                else:
                    warnings.warn('Cookie file not found.')
                return 0

            # Normal download mode:
            if args['user']:
                looter_cls = ProfileLooter
                target = args['<profile>']
            elif args['hashtag']:
                looter_cls = HashtagLooter
                target = args['<hashtag>']
            elif args['post']:
                looter_cls = PostLooter
                target = args['<post_token>']
            else:
                raise NotImplementedError("TODO")

            # Instantiate the looter
            looter = looter_cls(
                target,
                add_metadata=args['--add-metadata'],
                get_videos=args['--get-videos'],
                videos_only=args['--videos-only'],
                jobs=int(args['--jobs']) if args['--jobs'] is not None else 16,
                template=args['--template'],
                dump_json=args['--dump-json'],
                dump_only=args['--dump-only'],
                extended_dump=args['--extended-dump']
            )

            # Attempt to login and extract the timeframe
            if args['--username']:
                login(args)
            if args['--num-to-dl']:
                args['--num-to-dl'] = int(args['--num-to-dl'])
            try:
                if args['--time'] is not None:
                    args['--time'] = get_times_from_cli(args['--time'])
            except ValueError as ve:
                _print("invalid format for --time parameter:", args["--time"])
                _print("    (format is [D]:[D] where D is an ISO 8601 date)")
                return 1

            logger.debug("Opening destination filesystem")
            dest_url = args.get('<directory>') or os.getcwd()
            dest_fs = fs.open_fs(dest_url, create=True)

            logger.notice("Starting download of `%s`", target)
            n = looter.download(
                destination=dest_fs,
                media_count=args['--num-to-dl'],
                timeframe=args['--time'],
                new_only=args['--new'],
                pgpbar_cls=None if args['--quiet'] else TqdmProgressBar,
                dlpbar_cls=None if args['--quiet'] else TqdmProgressBar)
            if n > 1:
                logger.success("Downloaded %i posts.", n)
            elif n == 1:
                logger.success("Downloaded %i post.", n)

        except (Exception, KeyboardInterrupt) as e:
            from .threadutils import threads_force_join, threads_count
            # Show error traceback if any
            if not isinstance(e, KeyboardInterrupt):
                logger.critical("%s", e)
                if args["--traceback"]:
                    traceback.print_exc()
            else:
                logger.critical("Interrupted")
            # Close remaining threads spawned by InstaLooter.download
            count = threads_count()
            if count:
                logger.notice("Terminating %i remaining workers...", count)
                threads_force_join()
            # Return the error number if any
            errno = e.errno if hasattr(e, "errno") else None
            return errno if errno is not None else 1

        else:
            return 0

        finally:
            logger.debug("Closing destination filesystem")
            try:
                dest_fs.close()
            except Exception:
                pass



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

import unittest
import warnings

import fs.memoryfs
import parameterized
import requests
import six

from instalooter.looters import ProfileLooter, HashtagLooter, PostLooter

from .utils import mock
from .utils.method_names import signature


class TestInstaLooter(unittest.TestCase):

    MEDIA_COUNT = 5

    @classmethod
    def setUpClass(cls):
        cls.session = requests.Session()

    @classmethod
    def tearDownClass(cls):
        cls.session.close()

    def setUp(self):
        self.destfs = fs.memoryfs.MemoryFS()

    def tearDown(self):
        self.destfs.close()

    @parameterized.parameterized.expand([
        parameterized.param("instagram",),
        parameterized.param("instagram", get_videos=True),
        # parameterized.param("serotonine",),
    ], testcase_func_name=signature)
    def test_profile(self, profile, **kwargs):
        looter = ProfileLooter(profile, session=self.session, **kwargs)
        looter.download(self.destfs, media_count=self.MEDIA_COUNT)
        self.assertGreaterEqual(len(self.destfs.listdir("/")), self.MEDIA_COUNT)

    @parameterized.parameterized.expand([
        parameterized.param("eggs"),
        parameterized.param("python", videos_only=True),
    ], testcase_func_name=signature)
    def test_hashtag(self, hashtag, **kwargs):
        looter = HashtagLooter(hashtag, session=self.session, **kwargs)
        looter.download(self.destfs, media_count=self.MEDIA_COUNT)
        self.assertGreaterEqual(len(self.destfs.listdir("/")), self.MEDIA_COUNT)



class TestPostLooter(unittest.TestCase):

    @mock.patch('instalooter.looters.InstaLooter.__init__')
    def test_post_url(self, _):
        urls = (
            "http://www.instagram.com/p/BJlIB9WhdRn/?taken-by=2k",
            "https://www.instagram.com/p/BJlIB9WhdRn/?taken-by=2k",
            "www.instagram.com/p/BJlIB9WhdRn/?taken-by=2k",
            "http://instagr.am/p/BJlIB9WhdRn/?taken-by=2k",
            "https://instagr.am/p/BJlIB9WhdRn/?taken-by=2k",
            "instagr.am/p/BJlIB9WhdRn/?taken-by=2k",
        )
        for url in urls:
            looter = PostLooter(url)
            self.assertEqual(looter.code, "BJlIB9WhdRn")

    @mock.patch('instalooter.looters.InstaLooter.__init__')
    def test_invalid_post_code(self, _):
        with self.assertRaises(ValueError):
            looter = PostLooter("instagram")  # invalid code


# class TestTemplate(_TempTestCase):
#
#     MEDIA_COUNT = 30
#
#     def test_template_1(self):
#         profile = "therock"
#         looter = instaLooter.InstaLooter(
#             self.tmpdir, profile=profile, get_videos=True,
#             template='{username}-{id}'
#         )
#         looter.download(media_count=self.MEDIA_COUNT, with_pbar=False)
#         for f in os.listdir(self.tmpdir):
#             self.assertTrue(f.startswith(profile))
#
#
# class TestDump(_TempTestCase):
#
#     def assertMediaEqual(self, media, dump):
#         for key in ['__typename', 'date', 'dimensions', 'display_src',
#                     'is_video', 'media_preview']:
#             self.assertEqual(media[key], dump[key])
#
#         self.assertEqual(
#             media.get('code') or media['shortcode'],
#             dump.get('code' or dump['shortcode'])
#         )
#         self.assertEqual(
#             media['owner']['id'],
#             dump['owner']['id']
#         )
#         self.assertIn('likes', dump)
#         self.assertIn('comments', dump)
#
#     def test_dump_json(self):
#         looter = instaLooter.InstaLooter(
#             self.tmpdir,
#             profile="instagram",
#             dump_json=True,
#         )
#         test_medias = list(itertools.islice(
#             (m for m in looter.medias() if not m['is_video']), 3))
#         looter.download(media_count=3)
#
#         # Check all files were downloaded as expected
#         self.assertEqual(
#             sorted(os.listdir(self.tmpdir)),
#             sorted(f for media in test_medias for f in (
#                 str("{}.jpg").format(media['id']),
#                 str("{}.json").format(media['id']),
#             ))
#         )
#
#         # Check the metadata are OK
#         for media in test_medias:
#             with open(os.path.join(self.tmpdir, "{}.json").format(media['id'])) as f:
#                 dump = json.load(f)
#             self.assertMediaEqual(media, dump)
#
#     def test_dump_only(self):
#         looter = instaLooter.InstaLooter(
#             self.tmpdir,
#             profile="instagram",
#             dump_only=True,
#         )
#         test_medias = list(itertools.islice(
#             (m for m in looter.medias() if not m['is_video']), 3))
#         looter.download(media_count=3)
#
#         # Check all files were downloaded as expected
#         self.assertEqual(
#             sorted(os.listdir(self.tmpdir)),
#             sorted(str("{}.json").format(media['id']) for media in test_medias)
#         )
#
#         # Check the metadata are OK
#         for media in test_medias:
#             with open(os.path.join(self.tmpdir, "{}.json").format(media['id'])) as f:
#                 dump = json.load(f)
#             self.assertMediaEqual(media, dump)
#
#     def test_extended_dump(self):
#         looter = instaLooter.InstaLooter(
#             self.tmpdir,
#             profile="instagram",
#             dump_only=True,
#             extended_dump=True,
#         )
#         test_medias = list(itertools.islice(
#             (m for m in looter.medias() if not m['is_video']), 3))
#         looter.download(media_count=3)
#
#         # Check all files were downloaded as expected
#         self.assertEqual(
#             sorted(os.listdir(self.tmpdir)),
#             sorted(str("{}.json").format(media['id']) for media in test_medias)
#         )
#
#         # Check the metadata are OK
#         for media in test_medias:
#             with open(os.path.join(self.tmpdir, "{}.json").format(media['id'])) as f:
#                 dump = json.load(f)
#             self.assertMediaEqual(media, dump)
#
#             # Check the dump was "extended"
#             self.assertIn('edge_media_to_comment', dump)
#             self.assertIn('edge_media_to_caption', dump)
#
#
# class TestUtils(_TempTestCase):
#
#     MEDIA_COUNT = 30
#
#     def setUp(self):
#         super(TestUtils, self).setUp()
#         self.looter = instaLooter.InstaLooter()
#
#     def test_extract_post_code_from_url(self):
#         url = "https://www.instagram.com/p/BFB6znLg5s1/"
#
#         self.assertEqual(
#             self.looter._extract_code_from_url(url),
#             'BFB6znLg5s1',
#         )
#
#         with self.assertRaises(ValueError):
#             self.looter._extract_code_from_url(
#                 'https://www.instagram.com/'
#             )
#
#     def test_get_owner_info(self):
#         therock = self.looter.get_owner_info("BTHqEhWFR4y")
#         self.assertEqual(therock['username'], 'therock')
#         self.assertEqual(therock['id'], '232192182')
#         self.assertFalse(therock['is_private'])
#
#         gearbox = self.looter.get_owner_info("BfMWE3aFsEh")
#         self.assertEqual(gearbox['username'], 'gearboxsoftware')
#         self.assertEqual(gearbox['id'], '1409542965')
#         self.assertFalse(gearbox['is_private'])
#
#     def test_url_generator_nocallable(self):
#         with self.assertRaises(ValueError):
#             self.looter = instaLooter.InstaLooter(
#                 self.tmpdir, profile="instagram", url_generator=1
#             )
#
#     @unittest.skipIf(sys.version_info < (3,4),
#                      "operator.length_hint is a 3.4+ feature.")
#     def test_length_hint_empty(self):
#
#         looter = instaLooter.InstaLooter(profile="jkshksjdhfjkhdkfhk")
#         self.assertEqual(operator.length_hint(looter), 0)
#
#         looter = instaLooter.InstaLooter(hashtag="jkshksjdhfjkhdkfhk")
#         self.assertEqual(operator.length_hint(looter), 0)
#
#     @unittest.skipIf(sys.version_info < (3,4),
#                      "operator.length_hint is a 3.4+ feature.")
#     def test_length_hint(self):
#
#         looter = instaLooter.InstaLooter(self.tmpdir, profile="tide")
#         hint = operator.length_hint(looter)
#
#         # Check the post count is greater than 0
#         self.assertGreater(hint, 0)
#
#         # Download pictures and check if the count
#         # match (at most as many posts downloaded)
#         looter.download()
#         self.assertLessEqual(len(os.listdir(self.tmpdir)), hint)


# def load_tests(loader, tests, pattern):
#     suite = unittest.TestSuite()
#     TestProfileLooter.register_tests()
#     suite.addTests(loader.loadTestsFromTestCase(TestProfileLooter))
#     # suite.addTests(loader.loadTestsFromTestCase(TestHashtagDownload))
#     # suite.addTests(loader.loadTestsFromTestCase(TestTemplate))
#     return suite


def setUpModule():
   warnings.simplefilter('ignore')


def tearDownModule():
   warnings.simplefilter(warnings.defaultaction)



# coding: utf-8
"""Iterators over Instagram medias.

Iterators defined in this module wrap `PageIterator` instances to yield
individual medias defined in each page instead of whole pages.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import datetime
import typing

import six

if typing.TYPE_CHECKING:
    from typing import Any, Dict, List, Optional, Iterable, Set, Text


_I = typing.TypeVar('_I', bound='MediasIterator')


__all__ = [
    "MediasIterator",
    "TimedMediasIterator",
]


class MediasIterator(typing.Iterator[typing.Dict[typing.Text, typing.Any]]):
    """An iterator over the medias obtained from a page iterator.
    """

    def __init__(self, page_iterator):
        # type: (Iterable[Dict[Text, Any]]) -> None
        self._it = iter(page_iterator)
        self._seen = set()          # type: Set[Text]
        self._edges = []            # type: List[Dict[Text, Dict[Text, Any]]]
        self._finished = False
        self._total = None          # type: Optional[int]
        self._done = 0

    def __iter__(self):
        # type: (_I) -> _I
        return self

    def _next_page(self):
        # type: () -> Dict[Text, Any]
        data = next(self._it)
        section = next(s for s in six.iterkeys(data) if s.endswith('_media'))
        return data[section]

    def __next__(self):
        # type: () -> Dict[Text, Any]
        if self._finished:
            raise StopIteration

        if not self._edges:
            page = self._next_page()
            self._total = page['count']
            self._edges.extend(page['edges'])
            if not page['edges']:
                raise StopIteration

        media = self._edges.pop(0)
        self._done += 1

        if media['node']['id'] in self._seen:
            self._finished = True

        self._seen.add(media['node']['id'])
        return media['node']

    def __length_hint__(self):
        if self._total is None:
            try:
                page = self._next_page()
                self._total = page['count']
                self._edges.extend(page['edges'])
            except StopIteration:
                self._total = 0
        return self._total - self._done

    if six.PY2:
        next = __next__


class TimedMediasIterator(MediasIterator):
    """An iterator over the medias within a specific timeframe.
    """

    @staticmethod
    def get_times(timeframe):
        if timeframe is None:
            timeframe = (None, None)
        try:
            start_time = timeframe[0] or datetime.date.today()
            end_time = timeframe[1] or datetime.date.fromtimestamp(0)
        except (IndexError, AttributeError):
            raise TypeError("'timeframe' must be a couple of dates!")
        return start_time, end_time

    def __init__(self, page_iterator, timeframe=None):
        super(TimedMediasIterator, self).__init__(page_iterator)
        self.start_time, self.end_time = self.get_times(timeframe)

    def __next__(self):
        while True:
            media = super(TimedMediasIterator, self).__next__()
            timestamp = media.get('taken_at_timestamp') or media['date']
            media_date = datetime.date.fromtimestamp(timestamp)

            if self.start_time >= media_date >= self.end_time:
                return media
            elif media_date < self.end_time:
                self._finished = True
                raise StopIteration

    if six.PY2:
        next = __next__



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

import os
import unittest

import fs.memoryfs

from instalooter.looters import ProfileLooter


USERNAME = os.getenv("IG_USERNAME")
PASSWORD = os.getenv("IG_PASSWORD")


@unittest.skipIf(os.getenv("CI") == "true", "not working in CI")
@unittest.skipUnless(USERNAME and PASSWORD, "credentials required")
class TestLogin(unittest.TestCase):

    def setUp(self):
        self.looter = ProfileLooter(USERNAME, template="test")
        self.destfs = fs.memoryfs.MemoryFS()

    def tearDown(self):
        self.destfs.close()

    def test_login(self):

        self.assertFalse(self.looter.logged_in())
        self.assertRaises(RuntimeError, self.looter.medias)
        self.assertFalse(self.looter._cachefs.exists(self.looter._COOKIE_FILE))

        try:
            self.looter.login(USERNAME, PASSWORD)
            self.assertTrue(self.looter.logged_in())
            self.assertTrue(self.looter._cachefs.exists(self.looter._COOKIE_FILE))
            self.assertTrue(next(self.looter.medias()))
        finally:
            self.looter.logout()
            self.assertFalse(self.looter._cachefs.exists(self.looter._COOKIE_FILE))


    def test_download(self):

        try:
            self.looter.login(USERNAME, PASSWORD)
            self.looter.download(self.destfs)
            self.assertTrue(self.destfs.exists('test.jpg'))
            self.assertEqual(self.destfs.getbytes('test.jpg')[6:10], b'JFIF')
        finally:
            self.looter.logout()



# coding: utf-8
"""Backports and alternative library implementations.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import typing

try:
    import simplejson as json
except ImportError:
    import json  # type: ignore

try:
    import PIL.Image
    import piexif
except ImportError:
    PIL = None
    piexif = None

try:
    from operator import length_hint
except ImportError:
    def length_hint(obj, default=0):
        # type: (typing.Any, int) -> int
        """Return an estimate of the number of items in obj.

        This is useful for presizing containers when building from an
        iterable.

        If the object supports len(), the result will be
        exact. Otherwise, it may over- or under-estimate by an
        arbitrary amount. The result will be an integer >= 0.

        See Also:
            `PEP 424 <https://www.python.org/dev/peps/pep-0424/>`_

        """
        try:
            return len(obj)
        except TypeError:
            try:
                get_hint = type(obj).__length_hint__
            except AttributeError:
                return default
            try:
                hint = get_hint(obj)
            except TypeError:
                return default
            if hint is NotImplemented:
                return default
            if not isinstance(hint, int):
                raise TypeError("Length hint must be an integer, not %r" %
                                type(hint))
            if hint < 0:
                raise ValueError("__length_hint__() should return >= 0")
            return hint


__all__ = ["PIL", "piexif", "json", "length_hint"]



# coding: utf-8
"""Background download thread.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import io
import operator
import threading
import time

import requests
import six
import tenacity

from ._impl import PIL, piexif, json


class InstaDownloader(threading.Thread):
    """The background InstaLooter worker class.
    """

    _tenacity_options = {
        "stop": tenacity.stop_after_attempt(5),
        "wait": tenacity.wait_exponential(1, 10),
    }

    def __init__(self,
                 queue,
                 destination,
                 namegen,
                 add_metadata=False,
                 dump_json=False,
                 dump_only=False,
                 pbar=None,
                 session=None):

        super(InstaDownloader, self).__init__()

        self.queue = queue
        self.destination = destination
        self.namegen = namegen
        self.session = session or requests.Session()
        self.pbar = pbar

        self.dump_only = dump_only
        self.dump_json = dump_json or dump_only
        self.add_metadata = add_metadata

        self._killed = False
        self._downloading = None

        retry = tenacity.retry(**self._tenacity_options)
        self._DOWNLOAD_METHODS = {
            "GraphImage": retry(self._download_image),
            "GraphVideo": retry(self._download_video),
            "GraphSidecar": self._download_sidecar,
        }

    def _download_image(self, media):
        url = media['display_url']
        filename = self.namegen.file(media)

        if self.destination.exists(filename):
            return

        # FIXME: find a way to remove failed temporary downloads
        with self.destination.open(filename, "wb") as f:
            with self.session.get(url) as res:
                f.write(res.content)
        self._set_time(media, filename)

    def _download_video(self, media):
        url = media['video_url']
        filename = self.namegen.file(media)

        if self.destination.exists(filename):
            return

        # FIXME: find a way to remove failed temporary downloads
        with self.destination.open(filename, "wb") as f:
            with self.session.get(url) as res:
                for chunk in res.iter_content(io.DEFAULT_BUFFER_SIZE):
                    f.write(chunk)
        self._set_time(media, filename)

    def _download_sidecar(self, media):
        edges = media.pop('edge_sidecar_to_children')['edges']
        for edge in six.moves.map(operator.itemgetter('node'), edges):
            for key, value in six.iteritems(media):
                edge.setdefault(key, value)
            self._DOWNLOAD_METHODS[edge['__typename']](edge)

    def _set_time(self, media, filename):
        details = {}
        details["modified"] = details["accessed"] = details["created"] = \
            media.get('taken_at_timestamp') or media['date']
        self.destination.setinfo(filename, {"details": details})

    def _dump(self, media):
        basename = self.namegen.base(media)
        filename = "{}.json".format(basename)
        mode = "w" if six.PY3 else "wb"
        with self.destination.open(filename, mode) as dest:
            json.dump(media, dest, indent=4, sort_keys=True)

    def run(self):
        while not self._killed:
            try:
                media = self.queue.get_nowait()

                # Received a poison pill: break the loop
                if media is None:
                    self._killed = True

                else:
                    # Download media
                    if not self.dump_only:
                        self._DOWNLOAD_METHODS[media["__typename"]](media)
                    # Dump JSON metadata if needed
                    if self.dump_json:
                        self._dump(media)
                    # Update progress bar if any
                    if self.pbar is not None and not self._killed:
                        with self.pbar.get_lock():
                            self.pbar.update()

                self.queue.task_done()

            except six.moves.queue.Empty:
                time.sleep(1)

    def terminate(self):
        self._killed = True



# coding: utf-8
"""Iterators over Instagram media pages.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import abc
import hashlib
import itertools
import math
import time
import typing

import six
from requests import Session

from ._impl import json
from ._utils import get_shared_data

if typing.TYPE_CHECKING:
    from typing import Any, Dict, Iterator, Iterable, Optional, Text


__all__ = [
    "PageIterator",
    "HashtagIterator",
    "ProfileIterator",
]


@six.add_metaclass(abc.ABCMeta)
class PageIterator(typing.Iterator[typing.Dict[typing.Text, typing.Any]]):
    """An abstract Instagram page iterator.
    """

    PAGE_SIZE = 50
    INTERVAL = 2

    _BASE_URL = "https://www.instagram.com/graphql/query/"
    _section_generic = NotImplemented    # type: Text
    _section_media = NotImplemented      # type: Text

    def __init__(self, session, rhx):
        # type: (Session) -> None
        self._finished = False
        self._cursor = None     # type: Optional[Text]
        self._current_page = 0
        self._data_it = iter(self._page_loader(session, rhx))

    @abc.abstractmethod
    def _getparams(self, cursor):
        # type: (Optional[Text]) -> Text
        return NotImplemented

    def _page_loader(self, session, rhx):
        # type: (Session) -> Iterable[Dict[Text, Dict[Text, Any]]]
        while True:
            # Cache cursor for later
            cursor = self._cursor
            # Query data
            try:
                # Prepare the query
                params = self._getparams(cursor)
                json_params = json.dumps(params, separators=(',', ':'))
                magic = "{}:{}".format(rhx, json_params)
                session.headers['x-instagram-gis'] = hashlib.md5(magic.encode('utf-8')).hexdigest()
                url = self._URL.format(json_params)
                # Query the server for data
                with session.get(url) as res:
                    self._last_page = data = res.json()
                # Yield that same data until cursor is updated
                while self._cursor == cursor:
                    yield data['data']
            except KeyError as e:
                if data.get('message') == 'rate limited':
                    raise RuntimeError("Query rate exceeded (wait before next run)")
                time.sleep(10)
            # Sleep before next query
            time.sleep(self.INTERVAL)

    def __length_hint__(self):
        # type: () -> int
        try:
            data = next(self._data_it)
            c = data[self._section_generic][self._section_media]['count']
            total = int(math.ceil(c / self.PAGE_SIZE))
        except (StopIteration, TypeError):
            total = 0
        return total - self._current_page

    def __iter__(self):
        return self

    def __next__(self):

        if self._finished:
            raise StopIteration

        data = next(self._data_it)

        try:
            media_info = data[self._section_generic][self._section_media]
        except (TypeError, KeyError):
            self._finished = True
            raise StopIteration

        if not media_info['page_info']['has_next_page']:
            self._finished = True
        elif not media_info['edges']:
            self._finished = True
            raise StopIteration
        else:
            self._cursor = media_info['page_info']['end_cursor']
            self._current_page += 1

        return data[self._section_generic]

    if six.PY2:
        next = __next__


class HashtagIterator(PageIterator):
    """An iterator over the pages refering to a specific hashtag.
    """

    _QUERY_ID = "17882293912014529"
    _URL = "{}?query_id={}&variables={{}}".format(PageIterator._BASE_URL, _QUERY_ID)
    _section_generic = "hashtag"
    _section_media = "edge_hashtag_to_media"

    def __init__(self, hashtag, session, rhx):
        super(HashtagIterator, self).__init__(session, rhx)
        self.hashtag = hashtag

    def _getparams(self, cursor):
        return {
            "tag_name": self.hashtag,
            "first": self.PAGE_SIZE,
            "after": cursor
        }


class ProfileIterator(PageIterator):
    """An iterator over the pages of a user profile.
    """

    _QUERY_HASH = "42323d64886122307be10013ad2dcc44"
    #_QUERY_HASH = "472f257a40c653c64c666ce877d59d2b"
    _URL = "{}?query_hash={}&variables={{}}".format(PageIterator._BASE_URL, _QUERY_HASH)
    _section_generic = "user"
    _section_media = "edge_owner_to_timeline_media"

    @classmethod
    def _user_data(cls, username, session):
        url = "https://www.instagram.com/{}/".format(username)
        try:
            with session.get(url) as res:
                return get_shared_data(res.text)
        except (ValueError, AttributeError):
            raise ValueError("user not found: '{}'".format(username))

    @classmethod
    def from_username(cls, username, session):
        user_data = cls._user_data(username, session)
        if 'ProfilePage' not in user_data['entry_data']:
            raise ValueError("user not found: '{}'".format(username))
        data = user_data['entry_data']['ProfilePage'][0]['graphql']['user']
        if data['is_private'] and not data['followed_by_viewer']:
            con_id = next((c.value for c in session.cookies if c.name == "ds_user_id"), None)
            if con_id != data['id']:
                raise RuntimeError("user '{}' is private".format(username))
        return cls(data['id'], session, user_data['rhx_gis'])

    def __init__(self, owner_id, session, rhx):
        super(ProfileIterator, self).__init__(session, rhx)
        self.owner_id = owner_id

    def _getparams(self, cursor):
        return {
            "id": self.owner_id,
            "first": self.PAGE_SIZE,
            "after": cursor,
        }



# coding: utf-8
from __future__ import absolute_import

import threading

from ..worker import InstaDownloader


def threads_force_join():
    for t in threading.enumerate():
        if isinstance(t, InstaDownloader):
            t.terminate()
            t.join()


def threads_count():
    return sum(isinstance(t, InstaDownloader) for t in threading.enumerate())



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

import os
import json

import fs.path
import instalooter.looters


_test_dir = os.path.abspath(os.path.join(__file__, ".."))
_url = "tar://{}/ig_mock.tar.gz".format(_test_dir)


def get_mock_fs():
    return fs.open_fs(_url)


class MockPages(object):

    def __init__(self, profile):
        self.profile = profile

    def __call__(self):
        with get_mock_fs() as mockfs:
            with mockfs.open("pages/{}".format(self.profile)) as f:
                return iter(json.load(f))


if __name__ == "__main__":

    with fs.open_fs(_test_dir) as test_fs:
        if test_fs.exists(fs.path.basename(_url)):
            test_fs.remove(fs.path.basename(_url))

    with fs.open_fs(_url, create=True) as mockfs:
        mockfs.makedir("pages", recreate=True)
        nintendo = instalooter.looters.ProfileLooter("nintendo")
        with mockfs.open("pages/nintendo", "w") as f:
            json.dump(list(nintendo.pages()), f)

        fluoxetine = instalooter.looters.HashtagLooter("fluoxetine")
        with mockfs.open("pages/fluoxetine", "w") as f:
            pages_it = fluoxetine.pages_it
            json.dump([next(pages_it) for _ in range(3)], f)



# coding: utf-8
"""Run several jobs sharing a session using a configuration file.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import io
import getpass
import logging
import typing

import six
import verboselogs
from requests import Session

from .looters import HashtagLooter, ProfileLooter
from .pbar import TqdmProgressBar

if typing.TYPE_CHECKING:
    from typing import Any, Dict, Mapping, Optional, Text, Type, Union
    from .looter import InstaLooter


#: The module logger
logger = verboselogs.VerboseLogger(__name__)


class BatchRunner(object):
    """Run ``InstaLooter`` in batch mode, using a configuration file.
    """

    _CLS_MAP = {
        'users': ProfileLooter,
        'hashtag': HashtagLooter,
    }  # type: Mapping[Text, Type[InstaLooter]]

    def __init__(self, handle, args=None):
        # type: (Any, Optional[Mapping[Text, Any]]) -> None

        close_handle = False
        if isinstance(handle, six.binary_type):
            handle = handle.decode('utf-8')
        if isinstance(handle, six.text_type):
            _handle = open(handle)  # type: typing.IO
            close_handle = True
        else:
            _handle = handle

        try:
            self.args = args or {}
            self.parser = six.moves.configparser.ConfigParser()
            getattr(self.parser, "readfp" if six.PY2 else "read_file")(_handle)
        finally:
            if close_handle:
                _handle.close()

    @typing.overload
    def _getboolean(self, section_id, key, default):
        # type: (Text, Text, bool) -> bool
        pass

    @typing.overload
    def _getboolean(self, section_id, key):
        # type: (Text, Text) -> Optional[bool]
        pass

    @typing.overload
    def _getboolean(self, section_id, key, default):
        # type: (Text, Text, None) -> Optional[bool]
        pass

    def _getboolean(self, section_id, key, default=None):
        # type: (Text, Text, Optional[bool]) -> Optional[bool]
        if self.parser.has_option(section_id, key):
            return self.parser.getboolean(section_id, key)
        return default

    @typing.overload
    def _getint(self, section_id, key, default):
        # type: (Text, Text, None) -> Optional[int]
        pass

    @typing.overload
    def _getint(self, section_id, key):
        # type: (Text, Text) -> Optional[int]
        pass

    @typing.overload
    def _getint(self, section_id, key, default):
        # type: (Text, Text, int) -> int
        pass

    def _getint(self, section_id, key, default=None):
        # type: (Text, Text, Optional[int]) -> Optional[int]
        if self.parser.has_option(section_id, key):
            return self.parser.getint(section_id, key)
        return default

    @typing.overload
    def _get(self, section_id, key, default):
        # type: (Text, Text, None) -> Optional[Text]
        pass

    @typing.overload
    def _get(self, section_id, key):
        # type: (Text, Text) -> Optional[Text]
        pass

    @typing.overload
    def _get(self, section_id, key, default):
        # type: (Text, Text, Text) -> Text
        pass

    def _get(self, section_id, key, default=None):
        # type: (Text, Text, Optional[Text]) -> Optional[Text]
        if self.parser.has_option(section_id, key):
            return self.parser.get(section_id, key)
        return default

    def run_all(self):
        # type: () -> None
        """Run all the jobs specified in the configuration file.
        """
        logger.debug("Creating batch session")
        session = Session()

        for section_id in self.parser.sections():
            self.run_job(section_id, session=session)

    def run_job(self, section_id, session=None):
        # type: (Text, Optional[Session]) -> None
        """Run a job as described in the section named ``section_id``.

        Raises:
            KeyError: when the section could not be found.

        """
        if not self.parser.has_section(section_id):
            raise KeyError('section not found: {}'.format(section_id))

        session = session or Session()

        for name, looter_cls in six.iteritems(self._CLS_MAP):

                targets = self.get_targets(self._get(section_id, name))
                quiet = self._getboolean(
                    section_id, "quiet", self.args.get("--quiet", False))

                if targets:
                    logger.info("Launching {} job for section {}".format(name, section_id))

                for target, directory in six.iteritems(targets):
                    try:
                        logger.info("Downloading {} to {}".format(target, directory))
                        looter = looter_cls(
                            target,
                            add_metadata=self._getboolean(section_id, 'add-metadata', False),
                            get_videos=self._getboolean(section_id, 'get-videos', False),
                            videos_only=self._getboolean(section_id, 'videos-only', False),
                            jobs=self._getint(section_id, 'jobs', 16),
                            template=self._get(section_id, 'template', '{id}'),
                            dump_json=self._getboolean(section_id, 'dump-json', False),
                            dump_only=self._getboolean(section_id, 'dump-only', False),
                            extended_dump=self._getboolean(section_id, 'extended-dump', False),
                            session=session)

                        if self.parser.has_option(section_id, 'username'):
                            looter.logout()
                            username = self._get(section_id, 'username')
                            password = self._get(section_id, 'password') or \
                                getpass.getpass('Password for "{}": '.format(username))
                            looter.login(username, password)

                        n = looter.download(
                            directory,
                            media_count=self._getint(section_id, 'num-to-dl'),
                            # FIXME: timeframe=self._get(section_id, 'timeframe'),
                            new_only=self._getboolean(section_id, 'new', False),
                            pgpbar_cls=None if quiet else TqdmProgressBar,
                            dlpbar_cls=None if quiet else TqdmProgressBar)

                        logger.success("Downloaded %i medias !", n)

                    except Exception as exception:
                        logger.error(six.text_type(exception))

    def get_targets(self, raw_string):
        # type: (Optional[Text]) -> Dict[Text, Text]
        """Extract targets from a string in 'key: value' format.
        """
        targets = {}
        if raw_string is not None:
            for line in raw_string.splitlines():
                if line:
                    target, directory = line.split(':', 1)
                    targets[target.strip()] = directory.strip()
        return targets



# coding: utf-8
"""Instagram looters implementations.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import abc
import atexit
import copy
import functools
import random
import re
import threading
import time
import typing
import warnings

import fake_useragent
import fs
import six
from requests import Session
from six.moves.queue import Queue
from six.moves.http_cookiejar import FileCookieJar, LWPCookieJar

from . import __author__, __name__ as __appname__, __version__
from ._impl import length_hint, json
from ._utils import NameGenerator, CachedClassProperty, get_shared_data
from .medias import TimedMediasIterator, MediasIterator
from .pages import ProfileIterator, HashtagIterator
from .pbar import ProgressBar
from .worker import InstaDownloader

if typing.TYPE_CHECKING:
    from datetime import datetime
    from typing import (
        Any, Callable, Dict, Iterator, Iterable, List,
        Optional, Text, Tuple, Type, Union)
    from fs.base import FS
    from six.moves.http_cookiejar import CookieJar
    _T = typing.TypeVar("_T")
    _Timeframe = Tuple[Optional[datetime], Optional[datetime]]


__all__ = [
    "InstaLooter",
    "ProfileLooter",
    "HashtagLooter",
    "PostLooter",
]


@six.add_metaclass(abc.ABCMeta)
class InstaLooter(object):
    """A brutal Instagram looter that raids without API tokens.
    """

    @CachedClassProperty
    @classmethod
    def _cachefs(cls):
        """~fs.base.FS: the cache filesystem.
        """
        url = "usercache://{}:{}:{}".format(__appname__, __author__, __version__)
        return fs.open_fs(url, create=True)

    @CachedClassProperty
    @classmethod
    def _user_agents(cls):
        """~fake_useragent.UserAgent: a collection of fake user-agents.
        """
        filename = 'fake_useragent_{}.json'.format(fake_useragent.VERSION)
        return fake_useragent.UserAgent(
            path=cls._cachefs.getsyspath(filename),
            safe_attrs=['__name__', '__objclass__'])

    # str: The name of the cookie file in the cache filesystem
    _COOKIE_FILE = "cookies.txt"

    @classmethod
    def _init_session(cls, session=None):
        # type: (Optional[Session]) -> Session
        """Initialise the given session and load class cookies to its jar.

        Arguments:
            session (~requests.Session, optional): a `requests`
                session, or `None` to create a new one.

        Returns:
            ~requests.Session: an initialised session instance.

        """
        session = session or Session()
        # Load cookies
        session.cookies = LWPCookieJar(
            cls._cachefs.getsyspath(cls._COOKIE_FILE))
        try:
            typing.cast(FileCookieJar, session.cookies).load()
        except IOError:
            pass
        typing.cast(FileCookieJar, session.cookies).clear_expired_cookies()
        return session

    @classmethod
    def _login(cls, username, password, session=None):
        # type: (str, str, Optional[Session]) -> None
        """Login with provided credentials and session.

        Arguments:
            username (str): the username to log in with.
            password (str): the password to log in with.
            session (~requests.Session, optional): the session to use,
                or `None` to create a new session.

        Note:
            Code taken from LevPasha/instabot.py

        """
        session = cls._init_session(session)
        headers = copy.deepcopy(session.headers)
        homepage = "https://www.instagram.com/"
        login_url = "https://www.instagram.com/accounts/login/ajax/"
        data = {'username': username, 'password': password}

        try:
            session.headers.update({
                'Accept-Encoding': 'gzip, deflate',
                'Accept-Language': 'en-US,en;q=0.8',
                'Connection': 'keep-alive',
                'Content-Length': '0',
                'Host': 'www.instagram.com',
                'Origin': 'https://www.instagram.com',
                'Referer': 'https://www.instagram.com',
                'User-Agent': cls._user_agents.firefox,
                'X-Instagram-AJAX': '1',
                'X-Requested-With': 'XMLHttpRequest'
            })

            with session.get(homepage) as res:
                token = get_shared_data(res.text)['config']['csrf_token']
                session.headers.update({'X-CSRFToken': token})

            time.sleep(5 * random.random())  # nosec
            with session.post(login_url, data, allow_redirects=True) as login:
                token = next(c.value for c in login.cookies if c.name == 'csrftoken')
                session.headers.update({'X-CSRFToken': token})
                if not login.ok:
                    raise SystemError("Login error: check your connection")
                data = json.loads(login.text)
                if not data.get('authenticated', False):
                    raise ValueError('Login error: check your login data')

            time.sleep(5 * random.random())  # nosec
            with session.get(homepage) as res:
                if res.text.find(username) == -1:
                    raise ValueError('Login error: check your login data')
                try:
                    typing.cast(FileCookieJar, session.cookies).save()
                except IOError:
                    pass

        finally:
            session.headers = headers

    @classmethod
    def _logout(cls, session=None):
        # type: (Optional[Session]) -> None
        """Log out from current session.

        Also deletes the eventual cookie file left in the cache directory,
        to prevent new connections from using the old session ID.

        Arguments:
            session (~requests.Session): the session to use, or `None`
                to create a new session.

        Note:
            Code taken from LevPasha/instabot.py

        """
        session = cls._init_session(session)
        sessionid = cls._sessionid(session)
        if sessionid is not None:
            url = "https://www.instagram.com/accounts/logout/"
            session.post(url, data={"csrfmiddlewaretoken": sessionid})

        if cls._cachefs.exists(cls._COOKIE_FILE):
            cls._cachefs.remove(cls._COOKIE_FILE)

    @classmethod
    def _logged_in(cls, session=None):
        # type: (Optional[Session]) -> bool
        """Check if there is an open Instagram session.

        Arguments:
            session (~requests.Session): the session to use, or `None`
                to create a new session.

        Returns:
            bool: `True` if there's an active session, `False` otherwise.

        """
        return cls._sessionid(session) is not None

    @classmethod
    def _sessionid(cls, session=None):
        # type: (Optional[Session]) -> Optional[Text]
        """Get the ID of the currently opened Instagram session.

        Arguments:
            session (~requests.Session): the session to use, or `None`
                to create a new session.

        Returns:
            str or None: the session ID, if any, or `None`.

        """
        _session = cls._init_session(session)
        _cookies = typing.cast(FileCookieJar, _session.cookies)
        return next((ck.value for ck in _cookies
                     if ck.domain == ".instagram.com"
                     and ck.name == "ds_user_id"
                     and ck.path == "/"), None)

    def __init__(self,
                 add_metadata=False,    # type: bool
                 get_videos=False,      # type: bool
                 videos_only=False,     # type: bool
                 jobs=16,               # type: int
                 template="{id}",       # type: Text
                 dump_json=False,       # type: bool
                 dump_only=False,       # type: bool
                 extended_dump=False,   # type: bool
                 session=None           # type: Optional[Session]
                 ):
        # type: (...) -> None
        """Create a new looter instance.

        Arguments:
            add_metadata (bool): Add date and comment metadata to
                the downloaded pictures.
            get_videos (bool): Also get the videos from the given target.
            videos_only (bool): Only download videos (implies
                ``get_videos=True``).
            jobs (bool): the number of parallel threads to use to
                download media (12 or more is advised to have a true parallel
                download of media files).
            template (str): a filename format, in Python new-style-formatting
                format. See the the :ref:`Template` page of the documentation
                for available keys.
            dump_json (bool): Save each resource metadata to a
                JSON file next to the actual image/video.
            dump_only (bool): Only save metadata and discard the actual
                resource.
            extended_dump (bool): Attempt to fetch as much metadata as
                possible, at the cost of more time. Set to `True` if, for
                instance, you always want the top comments to be downloaded
                in the dump.
            session (~requests.Session or None): a `requests` session,
                or `None` to create a new one.

        """
        self.add_metadata = add_metadata
        self.get_videos = get_videos or videos_only
        self.videos_only = videos_only
        self.jobs = jobs
        self.namegen = NameGenerator(template)
        self.dump_only = dump_only
        self.dump_json = dump_json or dump_only
        self.extended_dump = extended_dump
        self.session = self._init_session(session)
        atexit.register(self.session.close)

        # Set a fake User-Agent
        if self.session.headers['User-Agent'].startswith('python-requests'):
            self.session.headers['User-Agent'] = self._user_agents.firefox

        # Get CSRFToken and RHX
        with self.session.get('https://www.instagram.com/') as res:
            token = get_shared_data(res.text)['config']['csrf_token']
            self.session.headers['X-CSRFToken'] = token
            self.rhx = get_shared_data(res.text)['rhx_gis']

    @abc.abstractmethod
    def pages(self):
        # type: () -> Iterator[Dict[Text, Any]]
        """Obtain an iterator over Instagram post pages.

        Returns:
            PageIterator: an iterator over the instagram post pages.

        """
        return NotImplemented

    def _medias(self,
                pages_iterator,     # type: Iterable[Dict[Text, Any]]
                timeframe=None      # type: Optional[_Timeframe]
                ):
        # type: (...) -> Iterator[Dict[Text, Any]]
        """Obtain an iterator over the medias of the given pages iterator.

        Arguments:
            pages_iterator (Iterator): an iterator over the Instagram
                pages, returned by `InstaLooter.pages`

        Returns:
            MediasIterator: an iterator over the medias in every pages.

        """
        if timeframe is not None:
            return TimedMediasIterator(pages_iterator, timeframe)
        return MediasIterator(pages_iterator)

    def medias(self, timeframe=None):
        # type: (Optional[_Timeframe]) -> Iterator[Dict[Text, Any]]
        """Obtain an iterator over the Instagram medias.

        Wraps the iterator returned by `InstaLooter.pages` to seamlessly
        iterate over the medias of all the pages.

        Returns:
            MediasIterator: an iterator over the medias in every pages.

        """
        return self._medias(self.pages(), timeframe)

    def get_post_info(self, code):
        # type: (str) -> dict
        """Get media information from a given post code.

        Arguments:
            code (str): the code of the post (can be obtained either
                from the ``shortcode`` attribute of media dictionaries, or
                from a post URL: ``https://www.instagram.com/p/<code>/``)

        Returns:
            dict: a media dictionaries, in the format used by Instagram.

        """
        url = "https://www.instagram.com/p/{}/".format(code)
        with self.session.get(url) as res:
            data = get_shared_data(res.text)
            return data['entry_data']['PostPage'][0]['graphql']['shortcode_media']

    def download_pictures(self,
                          destination,       # type: Union[str, fs.base.FS]
                          media_count=None,  # type: Optional[int]
                          timeframe=None,    # type: Optional[_Timeframe]
                          new_only=False,    # type: bool
                          pgpbar_cls=None,   # type: Optional[Type[ProgressBar]]
                          dlpbar_cls=None    # type: Optional[Type[ProgressBar]]
                          ):
        # type: (...) -> int
        """Download all the pictures to the provided destination.

        Actually a shortcut for `.download` with ``condition`` set
        to accept only images.

        """
        return self.download(
            destination,
            condition=lambda media: not media["is_video"],
            media_count=media_count,
            timeframe=timeframe,
            new_only=new_only,
            pgpbar_cls=pgpbar_cls,
            dlpbar_cls=dlpbar_cls,
        )

    def download_videos(self,
                        destination,       # type: Union[str, fs.base.FS]
                        media_count=None,  # type: Optional[int]
                        timeframe=None,    # type: Optional[_Timeframe]
                        new_only=False,    # type: bool
                        pgpbar_cls=None,   # type: Optional[Type[ProgressBar]]
                        dlpbar_cls=None,   # type: Optional[Type[ProgressBar]]
                        ):
        # type: (...) -> int
        """Download all videos to the provided destination.

        Actually a shortcut for `.download` with ``condition`` set
        to accept only videos.

        """
        return self.download(
            destination,
            condition=lambda media: media["is_video"],
            media_count=media_count,
            timeframe=timeframe,
            new_only=new_only,
            pgpbar_cls=pgpbar_cls,
            dlpbar_cls=dlpbar_cls,
        )

    def download(self,
                 destination,           # type: Union[str, fs.base.FS]
                 condition=None,        # type: Optional[Callable[[dict], bool]]
                 media_count=None,      # type: Optional[int]
                 timeframe=None,        # type: Optional[_Timeframe]
                 new_only=False,        # type: bool
                 pgpbar_cls=None,       # type: Optional[Type[ProgressBar]]
                 dlpbar_cls=None,       # type: Optional[Type[ProgressBar]]
                 ):
        # type: (...) -> int
        """Download all medias passing ``condition`` to destination.

        Arguments:
            destination (~fs.base.FS or str): the filesystem where to
                store the downloaded files, as a filesystem instance or
                FS URL.
            condition (function): the condition to filter the
                medias with. If `None` is given, a function is created using
                the ``get_videos`` and ``videos_only`` passed at object
                initialisation.
            media_count (int or None): the maximum number of medias
                to download. Leave to ``None`` to download everything from
                the target. *Note that more files can be downloaded, since
                a post with multiple images/videos is considered to be a
                single media*.
            timeframe (tuple or None): a tuple of two `~datetime.datetime`
                objects to enforce a time frame (the first item must be
                more recent). Leave to `None` to ignore times.
            new_only (bool): stop media discovery when already
                downloaded medias are encountered.
            pgpbar_cls (type or None): an optional `~.pbar.ProgressBar`
                subclass to use to display page scraping progress.
            dlpbar_cls (type or None): an optional `~.pbar.ProgressBar`
                subclass to use to display file download progress.

        Returns:
            int: the number of queued medias.

            May not be equal to the number of downloaded medias if some
            errors occurred during background download.

        """
        # Open the destination filesystem
        destination, close_destination = self._init_destfs(destination)

        # Create an iterator over the pages with an optional progress bar
        pages_iterator = self.pages()   # type: Iterable[Dict[Text, Any]]
        pages_iterator = pgpbar = self._init_pbar(pages_iterator, pgpbar_cls)

        # Create an iterator over the medias
        medias_iterator = self._medias(iter(pages_iterator), timeframe)

        # Create the media download bar from a dummy iterator
        dlpbar = self._init_pbar(
            six.moves.range(length_hint(medias_iterator)), dlpbar_cls)

        # Start a group of workers
        workers, queue = self._init_workers(
            dlpbar if dlpbar_cls is not None else None, destination)

        # Make sure exiting the main thread will shutdown workers
        atexit.register(self._shutdown_workers, workers)

        # Queue all medias
        medias_queued = self._fill_media_queue(
            queue, destination, medias_iterator, media_count,
            new_only, condition)

        # Once queuing the medias is fininished, finish the page progress bar
        # and set a new maximum on the download progress bar.
        if pgpbar_cls is not None:
            pgpbar.finish()                         # type: ignore
        if dlpbar_cls is not None:
            dlpbar.set_maximum(medias_queued)       # type: ignore

        # If no medias were queued, issue a warning
        # TODO: refine warning depending on download parameters
        if medias_queued == 0:
            warnings.warn("No medias found.")

        # Add poison pills to the queue and wait for workers to finish
        self._poison_workers(workers, queue)
        self._join_workers(workers, queue)

        # Once downloading is finished, finish the download progress bar
        # and close the destination if needed.
        if dlpbar_cls is not None:
            dlpbar.finish()                        # type: ignore
        if close_destination:
            destination.close()

        return medias_queued

    def login(self, username, password):
        # type: (str, str) -> None
        """Log the instance in using the given credentials.

        Arguments:
            username (str): the username to log in with.
            password (str): the password to log in with.

        """
        self._login(username, password, session=self.session)

    def logout(self):
        # type: () -> None
        """Log the instance out from the currently opened session.
        """
        self._logout(session=self.session)

    def logged_in(self):
        # type: () -> bool
        """Check if there's an open Instagram session.
        """
        return self._logged_in(self.session)

    def _init_pbar(self,
                   it,             # type: Iterable[_T]
                   pbar_cls=None,  # type: Optional[Type[ProgressBar]]
                   ):
        # type: (...) -> Iterable[_T]
        """Wrap an iterable within a `ProgressBar`.

        Arguments:
            it (~collections.Iterable): an iterable to wrap.
            pgpbar_cls (type or None): an optional `ProgressBar` subclass
                to use, or `None` to avoid using a progress bar.

        Returns:
            ~collections.Iterable: the wrapped iterable.

        """
        if pbar_cls is not None:
            if not issubclass(pbar_cls, ProgressBar):
                raise TypeError("pbar must implement the ProgressBar interface !")
            maximum = length_hint(it)
            it = pbar = pbar_cls(it)
            pbar.set_maximum(maximum)
            pbar.set_lock(threading.RLock())
        return it

    def _init_destfs(self, destination, create=True):
        # type: (Union[str, fs.base.FS], bool) -> Tuple[fs.base.FS, bool]
        """Open a filesystem either from a FS URL or filesystem instance.

        Arguments:
            destination (~fs.base.FS or str): the destination filesystem
                to open, as a filesystem instance or FS URL.
            create (bool): whether or not to create a new
                filesystem if it does not exist.

        Returns:
            (~fs.base.FS, bool): the open FS, and whether to close it.

        """
        close_destination = False
        if isinstance(destination, six.binary_type):
            destination = destination.decode('utf-8')
        if isinstance(destination, six.text_type):
            destination = fs.open_fs(destination, create=create)
            close_destination = True
        if not isinstance(destination, fs.base.FS):
            raise TypeError("<destination> must be a FS URL or FS instance.")
        return destination, close_destination

    def _fill_media_queue(self,
                          queue,            # type: Queue
                          destination,      # type: fs.base.FS
                          medias_iter,      # type: Iterable[Any]
                          media_count=None,  # type: Optional[int]
                          new_only=False,   # type: bool
                          condition=None,   # type: Optional[Callable[[dict], bool]]
                          ):
        # type: (...) -> int
        """Fill the download queue with medias from the provided iterator.

        Arguments:
            queue (~queue.Queue): the download queue to fill.
            destination (~fs.base.FS): the filesystem where to download
                the files.
            medias_iterator (~collections.Iterable): an iterable over the
                Instagram medias to download.
            media_count (int or None): the maximum number of new medias to
                download, or ``None`` to download all discoverable medias.
            new_only (bool): stop media discovery when a media that was
                already downloaded is encountered.
            condition (function or None): the condition to filter the medias
                with. If `None` is given, a function is created using the
                ``get_videos`` and ``videos_only`` passed at object
                initialisation.

        Returns:
            int: the number of queued medias.

            May not be equal to the number of downloaded medias if some
            errors occurred during downloads.

        """
        # Create a condition from parameters if needed
        if condition is not None:
            _condition = condition       # type: Callable[[dict], bool]
        else:
            if self.videos_only:
                def _condition(media): return media['is_video']
            elif not self.get_videos:
                def _condition(media): return not media['is_video']
            else:
                def _condition(media): return True

        # Queue all media filling the condition
        medias_queued = 0
        for media in six.moves.filter(_condition, medias_iter):

            # Check if the whole post info is required
            if self.namegen.needs_extended(media) or media["__typename"] != "GraphImage":
                media = self.get_post_info(media['shortcode'])

            # Check that sidecar children fit the condition
            if media['__typename'] == "GraphSidecar":
                # Check that each node fits the condition
                for sidecar in media['edge_sidecar_to_children']['edges'][:]:
                    if not _condition(sidecar['node']):
                        media['edge_sidecar_to_children']['edges'].remove(sidecar)

                # Check that the nodelist is not depleted
                if not media['edge_sidecar_to_children']['edges']:
                    continue

            # Check that the file does not exist
            # FIXME: not working well with sidecar
            if new_only and destination.exists(self.namegen.file(media)):
                break

            # Put the medias in the queue
            queue.put(media)
            medias_queued += 1

            if media_count is not None and medias_queued >= media_count:
                break

        return medias_queued

    # WORKERS UTILS

    def _init_workers(self,
                      pbar,         # type: Union[ProgressBar, Iterable, None]
                      destination,  # type: fs.base.FS
                      ):
        # type: (...) -> Tuple[List[InstaDownloader], Queue]

        workers = []        # type: List[InstaDownloader]
        queue = Queue()     # type: Queue

        for _ in six.moves.range(self.jobs):
            worker = InstaDownloader(
                queue=queue,
                destination=destination,
                namegen=self.namegen,
                add_metadata=self.add_metadata,
                dump_json=self.dump_json,
                dump_only=self.dump_only,
                pbar=pbar,
                session=self.session)
            worker.start()
            workers.append(worker)

        return workers, queue

    def _poison_workers(self, workers, queue):
        # type: (List[InstaDownloader], Queue) -> None
        for worker in workers:
            queue.put(None)

    def _join_workers(self, workers, queue):
        # type: (List[InstaDownloader], Queue) -> None
        if any(w.is_alive() for w in workers):
            for worker in workers:
                worker.join()

    def _shutdown_workers(self, workers):
        # type: (List[InstaDownloader]) -> None
        for worker in workers:
            worker.terminate()


class ProfileLooter(InstaLooter):
    """A looter targeting medias on a user profile.
    """

    def __init__(self, username, **kwargs):
        # type: (str, **Any) -> None
        """Create a new profile looter.

        Arguments:
            username (str): the username of the profile.

        See `InstaLooter.__init__` for more details about accepted
        keyword arguments.

        """
        super(ProfileLooter, self).__init__(**kwargs)
        self._username = username
        self._owner_id = None

    def pages(self):
        # type: () -> ProfileIterator
        """Obtain an iterator over Instagram post pages.

        Returns:
            PageIterator: an iterator over the instagram post pages.

        Raises:
            ValueError: when the requested user does not exist.
            RuntimeError: when the user is a private account
                and there is no logged user (or the logged user
                does not follow that account).

        """
        if self._owner_id is None:
            it = ProfileIterator.from_username(self._username, self.session)
            self._owner_id = it.owner_id
            return it
        return ProfileIterator(self._owner_id, self.session, self.rhx)


class HashtagLooter(InstaLooter):
    """A looter targeting medias tagged with a hashtag.
    """

    def __init__(self, hashtag, **kwargs):
        # type: (str, **Any) -> None
        """Create a new hashtag looter.

        Arguments:
            username (str): the hashtag to search for.

        See `InstaLooter.__init__` for more details about accepted
        keyword arguments.

        """
        super(HashtagLooter, self).__init__(**kwargs)
        self._hashtag = hashtag

    def pages(self):  # noqa: D102
        # type: () -> HashtagIterator
        return HashtagIterator(self._hashtag, self.session, self.rhx)


class PostLooter(InstaLooter):
    """A looter targeting a specific post.
    """

    _RX_URL = re.compile(
        r'(?:https?://)?(?:www\.instagram\.com|instagr\.am)/p/([0-9a-zA-Z_\-]{10,11})'
    )

    _RX_CODE = re.compile(
        r'^[0-9a-zA-Z_\-]{10,11}$'
    )

    def __init__(self, code, **kwargs):
        # type: (str, **Any) -> None
        """Create a new hashtag looter.

        Arguments:
            code (str): the code of the post to get.

        See `InstaLooter.__init__` for more details about accepted
        keyword arguments.

        """
        super(PostLooter, self).__init__(**kwargs)

        self._info = None   # type: Optional[dict]

        match = self._RX_URL.match(code)
        if match is not None:
            self.code = match.group(1)
        elif self._RX_CODE.match(code) is None:
            raise ValueError("invalid post code: '{}'".format(code))
        else:
            self.code = code

    @property
    def info(self):
        # type: () -> dict
        if self._info is None:
            self._info = self.get_post_info(self.code)
        return self._info

    def pages(self):
        # type: () -> Iterator[Dict[Text, Any]]
        """Return a generator that yields a page with only the refered post.

        Yields:
            dict: a page dictionary with only a single media.

        """
        yield {"edge_owner_to_timeline_media": {
            "count": 1,
            "page_info": {
                "has_next_page": False,
                "end_cursor": None,
            },
            "edges": [
                {"node": self.info}
            ],
        }}

    def medias(self, timeframe=None):
        """Return a generator that yields only the refered post.

        Yields:
            dict: a media dictionary obtained from the given post.

        Raises:
            StopIteration: if the post does not fit the timeframe.

        """
        info = self.info
        if timeframe is not None:
            start, end = TimedMediasIterator.get_times(timeframe)
            timestamp = info.get("taken_at_timestamp") or info["media"]
            if not (start >= timestamp >= end):
                raise StopIteration
        yield info

    def download(self,
                 destination,       # type: Union[str, fs.base.FS]
                 condition=None,    # type: Optional[Callable[[dict], bool]]
                 media_count=None,  # type: Optional[int]
                 timeframe=None,    # type: Optional[_Timeframe]
                 new_only=False,    # type: bool
                 pgpbar_cls=None,   # type: Optional[Type[ProgressBar]]
                 dlpbar_cls=None,   # type: Optional[Type[ProgressBar]]
                 ):
        # type: (...) -> int
        """Download the refered post to the destination.

        See `InstaLooter.download` for argument reference.

        Note:
            This function, opposed to other *looter* implementations, will
            not spawn new threads, but simply use the main thread to download
            the files.

            Since a worker is in charge of downloading a *media* at a time
            (and not a *file*), there would be no point in spawning more.

        """
        destination, close_destination = self._init_destfs(destination)

        queue = Queue()                           # type: Queue[Dict]
        medias_queued = self._fill_media_queue(
            queue, destination, iter(self.medias()), media_count,
            new_only, condition)
        queue.put(None)

        worker = InstaDownloader(
            queue=queue,
            destination=destination,
            namegen=self.namegen,
            add_metadata=self.add_metadata,
            dump_json=self.dump_json,
            dump_only=self.dump_only,
            pbar=None,
            session=self.session)
        worker.run()

        return medias_queued



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

import textwrap
import unittest
import warnings

import fs

from instalooter.cli import main
from instalooter.batch import BatchRunner


class TestBatchRunner(unittest.TestCase):

    def setUp(self):
        self.destfs = fs.open_fs("temp://")
        self.tmpdir = self.destfs.getsyspath("/")

    def tearDown(self):
        self.destfs.close()

    def test_cli(self):

        cfg = textwrap.dedent(
            """
            [my job]

            num-to-dl = 3
            quiet = true

            users:
                therock: {self.tmpdir}
                nintendo: {self.tmpdir}
            """
        ).format(self=self)

        with self.destfs.open('batch.ini', 'w') as batch_file:
            batch_file.write(cfg)

        retcode = main(["batch", self.destfs.getsyspath('batch.ini')])
        self.assertEqual(retcode, 0)
        self.assertGreaterEqual(
            len(list(self.destfs.filterdir("/", ["*.jpg"]))), 6)


def setUpModule():
   warnings.simplefilter('ignore')


def tearDownModule():
   warnings.simplefilter(warnings.defaultaction)



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

import datetime
import unittest
import json
import os

import contexter
import fs
import parameterized
import requests
import six
from six.moves.queue import Queue

from instalooter.cli import main
from instalooter.cli import time as timeutils
from instalooter.cli import threadutils
from instalooter.cli.constants import USAGE
from instalooter.cli.login import login
from instalooter.worker import InstaDownloader

from .utils import mock
from .utils.method_names import firstparam
from .utils.ig_mock import MockPages


class TestCLI(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.session = requests.Session()

    @classmethod
    def tearDownClass(cls):
        cls.session.close()

    def setUp(self):
        self.destfs = fs.open_fs("temp://")
        self.tmpdir = self.destfs.getsyspath("/")

    def tearDown(self):
        self.destfs.close()

    def test_user(self):
        with contexter.Contexter() as ctx:
            ctx << mock.patch('instalooter.cli.ProfileLooter.pages', MockPages('nintendo'))
            r = main(["user", "nintendo", self.tmpdir, "-q", '-n', '10'])
        self.assertEqual(r, 0)
        self.assertEqual(len(self.destfs.listdir('/')), 10)

    def test_single_post(self):
        r = main(["post", "BFB6znLg5s1", self.tmpdir, "-q"])
        self.assertEqual(r, 0)
        self.assertTrue(self.destfs.exists("1243533605591030581.jpg"))

    def test_dump_json(self):
        r = main(["post", "BIqZ8L8AHmH", self.tmpdir, '-q', '-d'])
        self.assertEqual(r, 0)

        self.assertTrue(self.destfs.exists("1308972728853756295.json"))
        self.assertTrue(self.destfs.exists("1308972728853756295.jpg"))

        with self.destfs.open("1308972728853756295.json") as fp:
            json_metadata = json.load(fp)

        self.assertEqual("1308972728853756295", json_metadata["id"])
        self.assertEqual("BIqZ8L8AHmH", json_metadata["shortcode"])

    def test_dump_only(self):
        r = main(["post", "BIqZ8L8AHmH", self.tmpdir, '-q', '-D'])
        self.assertEqual(r, 0)

        self.assertTrue(self.destfs.exists("1308972728853756295.json"))
        self.assertFalse(self.destfs.exists("1308972728853756295.jpg"))

        with self.destfs.open("1308972728853756295.json") as fp:
            json_metadata = json.load(fp)

        self.assertEqual("1308972728853756295", json_metadata["id"])
        self.assertEqual("BIqZ8L8AHmH", json_metadata["shortcode"])

    def test_usage(self):
        handle = six.moves.StringIO()
        main(["--usage"], stream=handle)
        self.assertEqual(handle.getvalue().strip(), USAGE.strip())

    def test_single_post_from_url(self):
        url = "https://www.instagram.com/p/BFB6znLg5s1/"
        main(["post", url, self.tmpdir, "-q"])
        self.assertIn("1243533605591030581.jpg", os.listdir(self.tmpdir))


class TestTimeUtils(unittest.TestCase):

    @parameterized.parameterized.expand([
        (":", (None, None)),
        ("2017-03-12:", (datetime.date(2017, 3, 12), None)),
        (":2016-08-04", (None, datetime.date(2016, 8, 4))),
        ("2017-03-01:2017-02-01", (datetime.date(2017, 3, 1), datetime.date(2017, 2, 1))),
    ], testcase_func_name=firstparam)
    def test_get_times_from_cli(self, token, expected):
        self.assertEqual(timeutils.get_times_from_cli(token), expected)

    @parameterized.parameterized.expand([
        ("thisday", 0, 0),
        ("thisweek", 7, 7),
        ("thismonth", 28, 31),
        ("thisyear", 365, 366),
    ], testcase_func_name=firstparam)
    def test_get_times_from_cli_keywords(self, token, inf, sup):
        start, stop = timeutils.get_times_from_cli(token)
        self.assertGreaterEqual(start - stop, datetime.timedelta(inf))
        self.assertLessEqual(start - stop, datetime.timedelta(sup))
        self.assertEqual(start, datetime.date.today())

    @parameterized.parameterized.expand([
        ["x"],
        ["x:y"],
        ["x:y:z"],
    ], testcase_func_name=firstparam)
    def test_get_times_from_cli_bad_format(self, token):
        self.assertRaises(ValueError, timeutils.get_times_from_cli, token)


@mock.patch('instalooter.looters.InstaLooter._login')
@mock.patch('getpass.getpass')
class TestLoginUtils(unittest.TestCase):

    def test_cli_login_no_username(self, getpass_, login_):
        args = {'--username': None, "--password": None}
        login(args)
        login_.assert_not_called()

    @mock.patch('instalooter.looters.InstaLooter._logged_in')
    def test_cli_login_no_password(self, logged_in_, getpass_, login_):
        args = {'--username': "user", "--password": None, "--quiet": False}
        logged_in_.return_value = False
        getpass_.return_value = "pasw"
        login(args)
        login_.assert_called_once_with("user", "pasw")

    @mock.patch('instalooter.looters.InstaLooter._logged_in')
    def test_cli_login(self, logged_in_, getpass_, login_):
        args = {'--username': "user", "--password": "pasw", "--quiet": False}
        logged_in_.return_value = False
        login(args)
        login_.assert_called_once_with("user", "pasw")

    @mock.patch('instalooter.looters.InstaLooter._logged_in')
    def test_cli_already_logged_in(self, logged_in_, getpass_, login_):
        args = {'--username': "user", "--password": "pasw", "--quiet": False}
        logged_in_.return_value = True
        login(args)
        login_.assert_not_called()


class TestThreadUtils(unittest.TestCase):

    def test_threads_count(self):

        q = Queue()
        t1 = InstaDownloader(q, None, None)
        t2 = InstaDownloader(q, None, None)

        try:
            self.assertEqual(threadutils.threads_count(), 0)
            t1.start()
            self.assertEqual(threadutils.threads_count(), 1)
            t2.start()
            self.assertEqual(threadutils.threads_count(), 2)
        finally:
            t1.terminate()
            t2.terminate()

    def test_threads_force_join(self):

        q = Queue()
        t1 = InstaDownloader(q, None, None)
        t2 = InstaDownloader(q, None, None)

        t1.start()
        t2.start()

        self.assertTrue(t1.is_alive())
        self.assertTrue(t2.is_alive())

        threadutils.threads_force_join()

        self.assertFalse(t1.is_alive())
        self.assertFalse(t2.is_alive())



# coding: utf-8
"""Internal utility classes and functions.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import datetime
import hashlib
import operator
import os
import re
import typing

import six

from ._impl import json

if typing.TYPE_CHECKING:
    from typing import Any, Dict, Mapping, Optional, Text


class NameGenerator(object):
    """Generator for filenames using a template.
    """

    @classmethod
    def _get_info(cls, media):
        # type: (Mapping[Text, Any]) -> Mapping[Text, Any]

        info = {
            'id': media['id'],
            'code': media['shortcode'],
            'ownerid': media['owner']['id'],
            'username': media['owner'].get('username'),
            'fullname': media['owner'].get('full_name'),
            'commentscount': media['edge_media_to_comment']['count'],
            'likescount': media['edge_media_preview_like']['count'],
            'width': media.get('dimensions', {}).get('width'),
            'height': media.get('dimensions', {}).get('height'),
        }  # type: Dict[Text, Any]

        timestamp = media.get('date') or media.get('taken_at_timestamp')
        if timestamp is not None:
            dt = datetime.datetime.fromtimestamp(timestamp)
            info['datetime'] = ("{0.year}-{0.month:02d}-{0.day:02d} {0.hour:02d}"
                "h{0.minute:02d}m{0.second:02d}s{0.microsecond}").format(dt)
            info['date'] = datetime.date.fromtimestamp(timestamp)

        return dict(six.moves.filter(
            operator.itemgetter(1), six.iteritems(info)))

    def __init__(self, template="{id}"):
        # type: (Text) -> None
        self.template = template

    def base(self, media):
        # type: (Mapping[Text, Any]) -> Text
        info = self._get_info(media)
        return self.template.format(**info)

    def file(self, media, ext=None):
        # type: (Mapping[Text, Any], Optional[Text]) -> Text
        ext = ext or ("mp4" if media['is_video'] else "jpg")
        return os.path.extsep.join([self.base(media), ext])

    def needs_extended(self, media):
        # type: (Mapping[Text, Any]) -> bool
        try:
            self.base(media)
            return False
        except KeyError:
            return True


class CachedClassProperty(object):

    def __init__(self, factory):
        if not isinstance(factory, (classmethod, staticmethod)):
            factory = classmethod(factory)
        self.factory = factory
        self.value = self.sentinel = object()

    def __get__(self, obj, klass=None):
        if self.value is self.sentinel:
            self.value = self.factory.__get__(obj, klass)()
        return self.value

    def __set__(self, obj, value):
        raise AttributeError("can't set attribute")


def get_shared_data(html):
    match = re.search(r'window._sharedData = ({[^\n]*});', html)
    return json.loads(match.group(1))



# coding: utf-8
from __future__ import absolute_import

def signature(func, param_num, params):
    args = ','.join("{!r}".format(a) for a in params.args)
    kwargs = ','.join("{}={!r}".format(k, v) for k,v in params.kwargs.items())
    if args and kwargs:
        return "{}({},{})".format(func.__name__, args, kwargs)
    else:
        return "{}({})".format(func.__name__, args or kwargs)

def firstparam(func, param_num, params):
    return "{}({!r})".format(func.__name__, params.args[0])

def num(func, param_num, params):
    return "{}_{}".format(func.__name__, param_num)



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

import datetime

import dateutil.relativedelta


def date_from_isoformat(isoformat_date):
    """Convert an ISO-8601 date into a `datetime.date` object.

    Argument:
        isoformat_date (str): a date in ISO-8601 format (YYYY-MM-DD)

    Returns:
        ~datetime.date: the object corresponding to the given ISO date.

    Raises:
        ValueError: when the date could not be converted successfully.

    See Also:
        `ISO-8601 specification <https://en.wikipedia.org/wiki/ISO_8601>`_.
    """
    year, month, day = isoformat_date.split('-')
    return datetime.date(int(year), int(month), int(day))


def get_times_from_cli(cli_token):
    """Convert a CLI token to a datetime tuple.

    Argument:
        cli_token (str): an isoformat datetime token ([ISO date]:[ISO date])
            or a special value among:
                * thisday
                * thisweek
                * thismonth
                * thisyear

    Returns:
        tuple: a datetime.date objects couple, where the first item is
            the start of a time frame and the second item the end of the
            time frame. Both elements can also be None, if no date was
            provided.

    Raises:
        ValueError: when the CLI token is not in the right format
            (no colon in the token, not one of the special values, dates
            are not in proper ISO-8601 format.)

    See Also:
        `ISO-8601 specification <https://en.wikipedia.org/wiki/ISO_8601>`_.

    """
    today = datetime.date.today()

    if cli_token=="thisday":
        return today, today
    elif cli_token=="thisweek":
        return today, today - dateutil.relativedelta.relativedelta(days=7)
    elif cli_token=="thismonth":
        return today, today - dateutil.relativedelta.relativedelta(months=1)
    elif cli_token=="thisyear":
        return today, today - dateutil.relativedelta.relativedelta(years=1)
    else:
        try:
            start_date, stop_date = cli_token.split(':')
        except ValueError:
            raise ValueError("--time parameter must contain a colon (:)")
        if not start_date and not stop_date: # ':', no start date, no stop date
            return None, None
        try:
            start_date = date_from_isoformat(start_date) if start_date else None
            stop_date = date_from_isoformat(stop_date) if stop_date else None
        except ValueError:
            raise ValueError("--time parameter was not provided ISO formatted dates")
        return start_date, stop_date



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

import datetime
import json
import os
import textwrap
import time
import unittest
import warnings

import contexter
import fs
import requests
import six

from instalooter._impl import length_hint, piexif, PIL
from instalooter.batch import BatchRunner, logger as batch_logger
from instalooter.cli import main
from instalooter.looters import HashtagLooter, ProfileLooter, PostLooter

from .utils import mock
from .utils.ig_mock import MockPages

# @mock.patch('instalooter.looter.requests.Session', lambda: TestResolvedIssues.session)


class TestResolvedIssues(unittest.TestCase):

    if six.PY2:
        assertRegex = unittest.TestCase.assertRegexpMatches

    @classmethod
    def setUpClass(cls):
        cls.session = requests.Session()

    @classmethod
    def tearDownClass(cls):
        cls.session.close()

    def setUp(self):
        self.destfs = fs.open_fs("temp://")
        self.tmpdir = self.destfs.getsyspath("/")
        warnings._showwarning = warnings.showwarning

    def tearDown(self):
        self.destfs.close()
        warnings.showwarning = warnings._showwarning

    @unittest.expectedFailure
    @unittest.skipUnless(piexif, "piexif required for this test")
    def test_issue_009(self):
        """
        Thanks to @kurtmaia for reporting this bug.

        Checks that adding metadata to pictures downloaded from a hashtag
        works as well.
        """
        looter = HashtagLooter("fluoxetine", add_metadata=True, session=self.session)
        with contexter.Contexter() as ctx:
            ctx << mock.patch.object(looter, 'pages', MockPages('fluoxetine'))
            looter.download(self.destfs, media_count=10)
        for f in self.destfs.listdir("/"):
            exif = piexif.load(self.destfs.getbytes(f))
            self.assertTrue(exif['Exif'])  # Date & Caption
            self.assertTrue(exif['0th'])  # Image creator

    def test_issue_012(self):
        """Feature request by @paramjitrohit.

        Allows downloading pictures and videos only within a timeframe.
        """
        looter = ProfileLooter("nintendo", session=self.session)
        day = datetime.date(2018, 3, 16)
        with contexter.Contexter() as ctx:
            ctx << mock.patch.object(looter, 'pages', MockPages('nintendo'))
            medias_in_timeframe = list(looter.medias(timeframe=[day, day]))
        self.assertEqual(len(medias_in_timeframe), 2)

    def test_issue_019(self):
        """
        Thanks to @emijawdo for reporting this bug.

        Checks that instalooter does not crash when not given a destination
        directory and uses the current directory.
        """
        initial_dir = os.getcwd()
        os.chdir(self.tmpdir)

        try:
            with contexter.Contexter() as ctx:
                ctx << mock.patch('instalooter.looters.InstaLooter.pages', MockPages('nintendo'))
                main(["user", "nintendo", "-n", "3", "-q"])
                self.assertGreaterEqual(len(self.destfs.listdir("/")), 3)
        finally:
            os.chdir(initial_dir)

    def test_issue_014(self):
        """Feature request by @JFLarsen.

        Allows customizing filenames using a template following Python
        `.format()` minilanguage.
        """
        looter = ProfileLooter("nintendo", template="{username}.{id}", session=self.session)
        with contexter.Contexter() as ctx:
            ctx << mock.patch.object(looter, 'pages', MockPages('nintendo'))
            looter.download(self.destfs, media_count=5)
        for f in self.destfs.scandir("/"):
            self.assertTrue(f.name.startswith('nintendo.'))

    @unittest.skipIf(os.getenv("IG_USERNAME") is None, "need private user account")
    def test_issue_006(self):
        """
        Checks that instalooter does not iterate forever on a private
        profile.
        """
        with self.assertRaises(RuntimeError):
            username = os.getenv("IG_USERNAME")
            looter = ProfileLooter(username, session=self.session)
            looter.logout()
            next(looter.medias())

    def test_issue_015(self):
        """
        Feature request by @MohamedIM.

        Checks that videos are not downloaded several times if present
        already in the destination directory.
        """
        looter = ProfileLooter("nintendo", session=self.session)

        with contexter.Contexter() as ctx:
            ctx << mock.patch.object(looter, 'pages', MockPages('nintendo'))
            looter.download_videos(self.destfs, media_count=1)
            video_file = next(self.destfs.filterdir("/", ["*.mp4"]))
            mtime = self.destfs.getdetails(video_file.name).accessed
            looter.download_videos(self.destfs, media_count=1)
            self.assertEqual(mtime, self.destfs.getdetails(video_file.name).accessed)

    def test_issue_022(self):
        """
        Thanks to @kuchenmitsahne for reporting this bug.

        Checks that using ``{datetime}`` in the template does not put
        a Windows forbidden character in the filename.
        """
        FORBIDDEN = set('<>:"/\|?*')
        looter = ProfileLooter("nintendo", template="{datetime}", session=self.session)
        with contexter.Contexter() as ctx:
            ctx << mock.patch.object(looter, 'pages', MockPages('nintendo'))
            looter.download(self.destfs, media_count=5)
        for f in self.destfs.scandir("/"):
            self.assertFalse(FORBIDDEN.intersection(f.name))

    def test_issue_026(self):
        """
        Feature request by @verafide.

        Checks that pictures that are downloaded are not
        resized.
        """
        PostLooter("BO0XpEshejh", session=self.session).download(self.destfs)
        pic = PIL.Image.open(self.destfs.getsyspath("1419863760138791137.jpg"))
        self.assertEqual(pic.size, (525, 612))

    def test_issue_039(self):
        """
        Feature request by @verafide

        Checks that all pictures are downloaded from posts
        with more than one picture.
        """
        looter = PostLooter("BRHecUuFhPl", session=self.session)
        looter.download(self.destfs)
        self.assertEqual(
            set(self.destfs.listdir("/")),
            {
                "1461270165803344956.jpg",
                "1461270167497776767.jpg",
                "1461270174435133336.jpg",
                "1461270172581471925.jpg",
                "1461270181565655668.jpg",
            }
        )

    def test_issue_042(self):
        """
        Thanks to @MohamedIM for reporting this bug.

        Checks that a multipost is successfully downloaded from
        the CLI `post` option.
        """
        looter = PostLooter('BRW-j_dBI6F', get_videos=True, session=self.session)
        looter.download(self.destfs)
        self.assertEqual(
            set(self.destfs.listdir("/")),
            {
                '1465633492745668095.mp4',
                '1465633517836005761.mp4',
                '1465633541559037966.mp4',
                '1465633561523918792.mp4',
            }
        )

    # OUTDATED: warn_windows is not used anymore
    #
    # def test_issue_044(self):
    #     """
    #     Thanks to @Bangaio64 for reporting this bug.
    #
    #     Checks that warn_windows does not trigger an exception.
    #     """
    #     import instalooter.utils
    #     warnings.showwarning = instalooter.utils.warn_windows
    #     looter = instalooter.InstaLooter(
    #         directory=self.tmpdir,
    #         profile="akjhdskjhfkjsdhfkjhdskjhfkjdshkfjhsdkjfdhkjdfshdfskhfd"
    #     )
    #     try:
    #         looter.download()
    #     except Exception:
    #         self.fail()

    def test_issue_041(self):
        """Feature request by @liorlior

        Allow downloading only videos.
        """
        looter = ProfileLooter("nintendo", videos_only=True, session=self.session)
        day = datetime.date(2017, 3, 10)
        with contexter.Contexter() as ctx:
            ctx << mock.patch.object(looter, 'pages', MockPages('nintendo'))
            looter.download(self.destfs, timeframe=[day, day])
        self.assertEqual(self.destfs.listdir("/"), ["1467639884243493431.mp4"])

    def test_issue_052(self):
        """Thanks to @cyrusclarke for reporting this bug.

        Checks that on hashtags with a lot of posts, the time parameter
        doesn't cause the program to crash without finding any media to
        download.
        """
        main(["hashtag", "happy", self.tmpdir, "-q", "-t", "thisweek", "-n", "5"])
        self.assertGreaterEqual(len(self.destfs.listdir('/')), 5)

    # OUTDATED: Sidecar info dicts are not converted anymore but passed
    #           to the workers directly.
    #
    # def test_issue_057(self):
    #     """
    #     Thanks to @VasiliPupkin256 for reporting this bug.
    #
    #     Checks that metadata can successfully extract caption
    #     out of multiposts containing images.
    #     """
    #     looter = ProfileLooter("awwwwshoot_ob", session=self.session)
    #     sidecar = next(m for m in looter.medias() if m['__typename'] == "GraphSidecar")
    #
    #     looter = PostLooter(sidecar['shortcode'], session=self.session)
    #     looter.download(self.destfs)
    #
    #     for key in ('caption', 'code', 'date'):
    #         self.assertIn(key, media)
    #         self.assertIsNotNone(media[key])

    def test_issue_066(self):
        """Thanks to @douglasrizzo for reporting this bug.

        Check that likescount and commentscount can be used
        in filename templates without causing the program to
        crash.
        """
        looter = ProfileLooter(
            "nintendo", get_videos=True, add_metadata=True,
            template='{id}-{likescount}-{commentscount}',
            session=self.session)
        with contexter.Contexter() as ctx:
            ctx << mock.patch.object(looter, 'pages', MockPages('nintendo'))
            looter.download(self.destfs, media_count=10)
        for image in self.destfs.listdir("/"):
            self.assertRegex(image, '[a-zA-Z0-9]*-[0-9]*-[0-9]*.(jpg|mp4)')

    def test_issue_076(self):
        """Thanks to @zeshuaro for reporting this bug.

        Check that when downloading hashtags, the downloader
        actually stops.
        """
        looter = HashtagLooter("oulianov", session=self.session)

        medias_it = looter.medias()
        postcount = length_hint(medias_it)

        for i, m in enumerate(medias_it):
            if i > postcount:
                self.fail("looter.medias() did not stop.")

    # OUTDATED: URLs are not modified anymore as Instagram prevents
    #           any modification
    #
    # def test_issue_082(self):
    #     """
    #     Thanks to @MohamedIM for reporting this bug.
    #
    #     Check that urls containing 'h-ak-igx' are not stripped from all
    #     their parameters.
    #     """
    #     looter = instalooter.looter.PostLooter('BWOYSYQDCo5', template='{code}')
    #     info = next(looter.medias())
    #
    #     info['display_url'] = \
    #         'https://ig-s-c-a.akamaihd.net/h-ak-igx/19764472_1586345694718446_4011887281420894208_n.jpg'
    #     looter.get_post_info = lambda code: info
    #
    #     looter.download_post('BWOYSYQDCo5')
    #
    #     with open(os.path.join(self.tmpdir, 'BWOYSYQDCo5.jpg'), 'rb') as f:
    #         self.assertNotIn(b'5xx Server Error', f.read())

    def test_issue_084(self):
        """Thanks to @raphaelbernardino for reporting this bug.

        Make sure private profiles with few pictures (less than a page worth)
        raise the private error as expected.
        """
        looter = ProfileLooter("rararudo", session=self.session)
        self.assertRaises(RuntimeError, looter.medias)

    @unittest.expectedFailure
    @unittest.skipUnless(piexif, "piexif required for this test")
    def test_issue_094(self):
        """Thanks to @jeanmarctst for raising this issue.

        Make sure caption is properly extracted from images downloaded
        from a post code and written to the metadata.
        """
        looter = PostLooter("BY77tSfBnRm",
                            add_metadata=True, template='{code}', session=self.session)
        looter.download(self.destfs)
        metadata = piexif.load(self.destfs.getbytes("BY77tSfBnRm.jpg"), True)
        self.assertTrue(metadata['Exif']['UserComment'])

    def test_issue_125(self):
        """Thanks to @applepanda for reporting this bug.

        Make sure colons in path do not cause issue in batch mode.
        """
        configfile = six.StringIO(textwrap.dedent(
            """
            [Family]
            users =
            	instagram: D:\\Instagram\\Profiles\\instagram
            	therock: D:\\Instagram\\Profiles\\therock
            """
        ))
        runner = BatchRunner(configfile)
        self.assertEqual(
            runner.get_targets(runner._get('Family', 'users')),
            {'instagram': 'D:\\Instagram\\Profiles\\instagram',
             'therock': 'D:\\Instagram\\Profiles\\therock'}
        )

    @mock.patch('instalooter.looters.InstaLooter.__init__')
    def test_issue_184(self, _):
        """Feature request by @ghost.

        Allow downloading a post directly from its URL.
        """
        looter = PostLooter("https://www.instagram.com/p/BJlIB9WhdRn/?taken-by=2k")
        self.assertEqual(looter.code, "BJlIB9WhdRn")

    def test_issue_185(self):
        """Feature request by @JPNYC81.

        Make sure an ``instalooter`` batch keeps even if it encounters errors
        on a specific job. This test tries with an non-existing profile.
        """
        configfile = six.StringIO(textwrap.dedent(
            """
            [Family]
            num-to-dl = 3
            users =
                jdskjhjkfhkdshfkjdhsfjsfdkjhfksdjhf: {tmp}
            	instagram: {tmp}
            	therock: {tmp}
            """
        ).format(tmp=self.tmpdir))
        runner = BatchRunner(configfile)
        with mock.patch('instalooter.batch.logger'):
            runner.run_all()
        self.assertGreaterEqual(len(self.destfs.listdir('/')), 6)

    def test_issue_194(self):
        """Feature request by @raphaelbernardino

        When trying to download from an non-existing user, try to display a
        meaningful message instead of a cryptic error.
        """
        username = "jdhfdjkhdlqdhfdhqfqjqlhfhdsdjquryerhdjfhqlkdfhkqhfqkure"
        looter = ProfileLooter(username)
        with self.assertRaises(ValueError) as ctx:
            media = next(looter.medias())
        self.assertEqual(str(ctx.exception), "user not found: '{}'".format(username))


# @mock.patch('instalooter.looter.requests.Session', lambda: TestPullRequests.session)
class TestPullRequests(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.session = requests.Session()

    @classmethod
    def tearDownClass(cls):
        cls.session.close()

    def setUp(self):
        self.destfs = fs.open_fs("temp://")
        self.tmpdir = self.destfs.getsyspath("/")

    def tearDown(self):
        self.destfs.close()

    def _pr_122_looter(self):
        return ProfileLooter('nintendo', template='{code}', session=self.session)

    def test_pr_122_download_post(self):
        """Feature implemented by @susundberg.

        Set the access time and modification time of a downloaded media
        according to its IG date.
        """
        code = 'BY77tSfBnRm'
        post_looter = PostLooter(code, session=self.session, template='{code}')
        info = post_looter.get_post_info(code)
        post_looter.download(self.destfs)
        stat = self.destfs.getdetails('{}.jpg'.format(code))
        self.assertEqual(stat.raw["details"]["accessed"], info['taken_at_timestamp'])
        self.assertEqual(stat.raw["details"]["modified"], info['taken_at_timestamp'])

    def test_pr_122_download_pictures(self):
        """Feature implemented by @susundberg.

        Set the access time and modification time of a downloaded media
        according to its IG date.
        """
        # Test download_pictures
        looter = self._pr_122_looter()
        with contexter.Contexter() as ctx:
            ctx << mock.patch.object(looter, 'pages', MockPages('nintendo'))
            pic = next(m for m in looter.medias() if not m['is_video'])
            looter.download_pictures(self.destfs, media_count=1)
        stat = self.destfs.getdetails('{}.jpg'.format(pic['shortcode']))
        self.assertEqual(stat.raw["details"]["accessed"], pic['taken_at_timestamp'])
        self.assertEqual(stat.raw["details"]["modified"], pic['taken_at_timestamp'])

    def test_pr_122_download_videos(self):
        """Feature implemented by @susundberg.

        Set the access time and modification time of a downloaded media
        according to its IG date.
        """
        # Test download_videos
        looter = self._pr_122_looter()
        with contexter.Contexter() as ctx:
            ctx << mock.patch.object(looter, 'pages', MockPages('nintendo'))
            vid = next(m for m in looter.medias() if m['is_video'])
            looter.download_videos(self.destfs, media_count=1)
        stat = self.destfs.getdetails('{}.mp4'.format(vid['shortcode']))
        self.assertEqual(stat.raw["details"]["accessed"], vid['taken_at_timestamp'])
        self.assertEqual(stat.raw["details"]["modified"], vid['taken_at_timestamp'])


def setUpModule():
   warnings.simplefilter('ignore')


def tearDownModule():
   warnings.simplefilter(warnings.defaultaction)



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

import textwrap


WARNING_ACTIONS = {'error', 'ignore', 'always', 'default', 'module', 'once'}


HELP = textwrap.dedent(
    """
    instalooter - Another API-less Instagram pictures and videos downloader

    Usage:
        instalooter (-h | --help | --version | --usage)
        instalooter batch <batch_file> [<directory>] [options]
        instalooter hashtag <hashtag> [<directory>] [options]
        instalooter user <profile> [<directory>] [options]
        instalooter post <post_token> [<directory>] [options]
        instalooter logout
        instalooter login [options]

    Arguments:
        <profile>                    The username of the profile to download
                                     videos and pictures from.
        <hashtag>                    A hashtag to download pictures and
                                     videos from.
        <post_token>                 Either the url or the code of a post to
                                     download the picture or video from.
        <directory>                  The directory in which to download files.
                                     Can actually be a Pyfilesystem2 FS URL
                                     (see http://pyfilesystem2.rtfd.io).
        <batch_file>                 The path to the batch file containing
                                     batch download instructions (see the
                                     online documentation).

    Options - Credentials:
        -u USER, --username USER     The username to connect to Instagram with.
        -p PASS, --password PASS     The password to connect to Instagram with
                                     (will be asked in the shell if the
                                     `--username`  option was given without
                                     the corresponding `--password`).

    Options - Files:
        -n NUM, --num-to-dl NUM      Maximum number of new files to download
        -j JOBS, --jobs JOBS         Number of parallel threads to use to
                                     download files. [default: 16]
        -T TMPL, --template TMPL     A filename template to use to write the
                                     files (see *Template*). [default: {id}]
        -v, --get-videos             Get videos as well as photos.
        -V, --videos-only            Get videos only. Implies `--get-videos`.
        -N, --new                    Only look for files newer than the ones
                                     in the destination directory (faster).
        -t TIME, --time TIME         The time limit within which to download
                                     pictures and video (see *Time*).

    Options - Metadata:
        -m, --add-metadata           Add date and caption metadata to downloaded
                                     pictures (requires PIL/Pillow and piexif).
        -d, --dump-json              Save metadata to a JSON file next to
                                     downloaded videos/pictures.
        -D, --dump-only              Save only the metadata and no video/picture.
                                     Implies `--dump-json`.
        -e, --extended-dump          Always dump the maximum amount of extracted
                                     information, at the cost of more time.

    Options - Miscellaneous:
        -l LEVEL, --loglevel LEVEL   The level of log to produce, as an
                                     integer or a level name. [default: INFO]
        -q, --quiet                  Do not display any output or progress
                                     bar. Implies `--loglevel ERROR`.
        -h, --help                   Display this message and quit.
        --version                    Show program version and quit.
        --traceback                  Print error traceback if any (use when
                                     reporting an issue on GitHub, please!).
        -W WARNINGCTL                Change warning behaviour (same as the
                                     Python `-W` flag). [default: default]

    Template:
        The default filename of the pictures and videos on Instagram doesn't
        show anything about the file you just downloaded. But using the -T
        argument allows you to give instalooter a filename template, using the
        the following format with brackets-enclosed ({}) variable names among:
        - ``id``*² and ``code``*² of the instagram id of the media
        - ``ownerid``*, ``username`` and ``fullname`` of the owner
        - ``datetime``*: the date and time of the post (YYYY-MM-DD hh:mm:ss)
        - ``date``*: the date of the post (YYYY-MM-DD)
        - ``width``* and ``height``*
        - ``likescount``* and ``commentscount``*

        ²: use at least one of these to make sure the generated file name
        is unique (``datetime`` is not unique anymore since multiple posts).

        *: use these only to quicken download, since fetching the others may
        take a tad longer (in particular in hashtag download mode).

        You are however to make sure that the generated filename is unique,
        so you should use at least id, code or datetime somewhere.
        Examples of acceptable values:
            - {username}.{datetime}.{code}
            - {username}-{likescount}-{width}x{height}.{id}

    Time:
        The --time parameter can be given either a combination of start and stop
        date in ISO format (e.g. 2016-12-21:2016-12-18, 2015-03-07:, :2016-08-02)
        or a special value among: "thisday", "thisweek", "thismonth", "thisyear".

        Edges are included in the time frame, so if using the following value:
        `--time 2016-05-10:2016-04-03`, then all medias will be downloaded
        including the ones posted the 10th of May 2016 and the 3rd of April 2016.

    See more at http://instalooter.readthedocs.io/en/latest/usage.html

    """
)


USAGE = next(s for s in HELP.split("\n\n") if s.startswith("Usage"))



# coding: utf-8
"""Progress bars used to report `InstaLooter.download` progress.

The module exposes and abstract class that can be derived to implement
your own progress displayer. The default implementation (which uses the
`tqdm` library) is used by the CLI.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import abc
import typing

import six
import tqdm

if typing.TYPE_CHECKING:
    from threading import Lock, RLock
    from typing import Union


_T = typing.TypeVar('_T', covariant=True)
_L = typing.TypeVar('_L')


@six.add_metaclass(abc.ABCMeta)
class ProgressBar(typing.Iterator[_T]):
    """An abstract progess bar used to report interal progress.
    """

    def __init__(self, it, *args, **kwargs):
        self.it = it
        self.__lock = None  # type: Union[Lock, RLock, None]

    def __iter__(self):
        # type: () -> ProgressBar[_T]
        return self

    def __next__(self):
        # type: () -> _T
        item = next(self.it)
        self.update()
        return item

    if six.PY2:
        next = __next__

    @abc.abstractmethod
    def update(self):
        # type: () -> None
        """Update the progress bar by one step.
        """
        return NotImplemented

    @abc.abstractmethod
    def set_maximum(self, maximum):
        # type: (int) -> None
        """Set the maximum number of steps of the operation.
        """
        return NotImplemented

    def finish(self):
        # type: () -> None
        """Notify the progress bar the operation is finished.
        """
        pass

    def set_lock(self, lock):
        # type: (Union[Lock, RLock]) -> None
        """Set a lock to be used by parallel workers.
        """
        self.__lock = lock

    def get_lock(self):
        # type: () -> Union[Lock, RLock]
        """Obtain the progress bar lock.
        """
        if self.__lock is None:
            raise RuntimeError("lock was not initialised")
        return self.__lock


class TqdmProgressBar(tqdm.tqdm, ProgressBar):
    """A progress bar using the `tqdm` library.
    """

    def __init__(self, it, *args, **kwargs):  # noqa: D102, D107
        kwargs["leave"] = False
        super(TqdmProgressBar, self).__init__(it, *args, **kwargs)
        ProgressBar.__init__(self, it)

    def set_maximum(self, maximum):  # noqa: D102
        self.total = maximum

    def update(self):
        """Update the progress bar by one step.
        """
        super(TqdmProgressBar, self).update()

    def finish(self):  # noqa: D102
        self.close()



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

import getpass
import logging

import verboselogs

from ..looters import InstaLooter


logger = verboselogs.VerboseLogger(__name__)


def login(args):
    if args['--username']:
        username = args['--username']
        if not InstaLooter._logged_in():
            password = args['--password'] or getpass.getpass()
            InstaLooter._login(username, password)
            if not args['--quiet']:
                logger.success('Logged in.')
        elif not args['--quiet']:
            logger.success("Already logged in.")
